#include "Controller.h"
#include <string>
#include <sstream>
#include <iostream>
#include <list>

#include <RL_PP/State.h>
#include <RL_PP/AbstractAction.h>
#include <RL_PP/ActionTileCoding.h>
#include <RL_PP/ActionTilings.h>
#include <RL_PP/Utility.h>

/*
Die Klasse Action ist die einfachste Implementation einer Aktion für die Aufgabe.
Möglich wäre z.B. die torque für das Kommando erst aus einem besseren/abstrakten
 Parameterzu berechnen.
*/
class Action : public AbstractAction
{
  public:
    Action( double t ) : AbstractAction()
    {
      torque = t;
    }

    inline const std::string command() const
    {
      std::ostringstream tmp;
      tmp << "(torque 2 " << torque << ")\n";
      return tmp.str();
    }

  private:
    double torque;
};



int Controller::theSocket;


Controller::Controller( const int aSocket)
  : Watkins( "rl.conf" )
{
  theSocket = aSocket;

  // will ich neu lernen -> loadValueFunction = false;
  // will ich das gelernte anwenden / weiterlernen -> true
  bool loadValueFunction = false;
  if ( loadValueFunction ) {
    load();
    // hier die Aktionen setzen
    avfPtr->setAction( 0, new Action( 0.0 ) );
    avfPtr->setAction( 1, new Action( 30.0 ) );
    avfPtr->setAction( 2, new Action( -30.0 ) );
  }
  else {
    State* statePtr = new State();
    // Zustandsvariablen mit MIN- und MAX-Wert registrieren
    posId = statePtr->registerVariable( "ANGLE", 0.0, 1024.0 );
  
    // ValueFunction anlegen
    // hier "ActionTileCoding"
    avfPtr = new ActionTileCoding( statePtr );
    // hier die Aktionen bei der valueFunction registrieren
    avfPtr->registerAction( new Action( 0.0 ) );
    avfPtr->registerAction( new Action( 30.0 ) );
    avfPtr->registerAction( new Action( -30.0 ) );
    // Zusandsraum festlegen
    // hier mit TileCodings
    Types::TILINGS_VAR_STRUCT vs;
    std::list< Types::TILINGS_VAR_STRUCT > vsList;
    vs.id = posId;
    vs.resolution = 5.0; // gewünschte Auflösung
    vs.tileType = Types::CLIP;
    vsList.push_back( vs );
    vsList.push_back( vs );
    ActionTilings* t = new ActionTilings( 4, vsList ); //Anzahl der Tiles
    static_cast< ActionTileCoding* >(avfPtr)->registerTiling( t );
  }
}


void Controller::init()
{
  recv ( theSocket, buffer, 1024, 0 );
  //alle joints in Ausgangslage fahren und fixieren
  sendMsg("(pidmaxtorque** 800)\n");
  sendMsg("(pid* 512 220 800 202 800 512 512 512 512 512 512 512 512 512 512 512 512 512 512)\n");
  sendMsg("(done)\n");
  for ( int i = 0; i <= 100; ++i ) {
    recv ( theSocket, buffer, 1024, 0 );
    sendMsg("(done)\n");
  }
  recv ( theSocket, buffer, 1024, 0 );
  // rechte Schulter lockern
  sendMsg("(pid 2 -1)\n");
  sendMsg("(done)\n");
  recv ( theSocket, buffer, 1024, 0 );
  // Reibung ausschalten
  sendMsg("(friction 2 0)\n");
  sendMsg("(done)\n");
  // etwas warten
  for ( int i = 0; i <= 3; ++i ) {
    recv ( theSocket, buffer, 1024, 0 );
    sendMsg("(done)\n");
  }
}


void Controller::run()
{
  // Aufruf der RL++-Kontrollschleife
  controlLoop();
}



void Controller::setInitialState()
{
  init();
  steps=0;
  //Umweltinformationen empfangen
  sprintf(buffer, "");
  int len = recv ( theSocket, buffer, 1024, 0 );
  buffer[len] = '\0';
  unsigned int p = 0;
  sscanf( buffer, "(time %*f)(angle %*i %i %*s", &p );
  // aktuellen Winkel bestimmen
  thePos = static_cast<double>(p);

  // neue Werte der Zustandsvariablen setzen
  avfPtr->getStatePtr()->setValueDbl( posId, thePos );

  std::cout << "pos = " << thePos << "\n";
  sendMsg("(done)\n");
}


void Controller::takeAction()
{
  const Action* action = static_cast< const Action* >( avfPtr->getActiveAction() );
  if ( action == NULL ) {
    std::cerr << "USERFEHLER: avfPtr->getActiveAction() liefert NULL !!!\n";
    exit(-1);
  }
  sendMsg(action->command().c_str());
  sendMsg("(done)\n");
}


const bool Controller::observeState()
{
  ++steps;
  sprintf(buffer, "");
  int len = recv ( theSocket, buffer, 1024, 0 );
  buffer[len] = '\0';
  unsigned int p = 0;
  sscanf( buffer, "(time %*f)(angle %*i %i %*s", &p );
  // aktuellen Winkel bestimmen
  thePos = static_cast<double>(p);

  // neue Werte der Zustandsvariablen setzen
  avfPtr->getStatePtr()->setValueDbl( posId, thePos );

  std::cout << "pos = " << thePos << "\n";

  if ( steps > 500 ) {
    sendMsg("(done)\n");
    return false;
  }
  return true;

}


const double Controller::calculateReward()
{
  double reward = 0.0;
  // Differenz zur Zielposition bestimmen
  double diff = fabs(202.0 - thePos);
  if ( diff > 512.0 ) {
    diff = 1023 - diff;
  }
  // Wenn Zielposition erreicht, gibt's Belohnung
  if (diff < 5.0) {
    reward = 1;
  }
  // sonst nicht
  return reward;
}


void Controller::finishEpisode()
{
  // valueFunction speichern
  save();
}


void Controller::save() const
{
  // save valuefunction
  std::cerr << "saving ... ";
  int res = avfPtr->save( "aufgabe3.vf" );
  if (res==0) {
    std::cerr << "okay\n";
  }
  else {
    std::cerr << "failed\n";
  }
}

void Controller::load()
{
  // load valuefunction
  avfPtr = new ActionTileCoding( "aufgabe3.vf" );
  avfPtr->setEpisodeStateHistory( true );
  posId = avfPtr->getStatePtr()->getVariableId( "ANGLE" );
}
